/**
 * Copyright Copyright 2007-8 Simon Andrews
 *
 *    This file is part of StackMeasure.
 *
 *    StackMeasure is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    StackMeasure is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with StackMeasure; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
import java.awt.Color;
import java.util.Vector;

public class SpatialGroup implements Comparable {

	private GroupCollection collection;
	private StackMeasureFile file;
	private Pixel [][][] outerPixels;
	private int startingHue;
	private Pixel [] pixels = null;
	private int pixelCount = 0;
	private int totalHueDeviation = 0;
	private int totalRed = 0;
	private int totalGreen = 0;
	private int totalBlue = 0;
	private Pixel brightestPixel = null;
		
	public SpatialGroup (Pixel p, GroupCollection collection) {
		file = p.file();
		outerPixels = new Pixel[collection.getFiles().getSlices()][collection.getFiles().getWidth()][2];
		this.collection = collection;
		outerPixels[p.slice()-1][p.x()][0] = p;
		outerPixels[p.slice()-1][p.x()][1] = p;
		pixelCount++;
		totalRed += p.r() & 0xff;
		totalGreen += p.g() & 0xff;
		totalBlue += p.b() & 0xff;
		totalHueDeviation = 0;
		brightestPixel = p;
		
		startingHue = p.h();
	}
	
	public boolean addPixel (Pixel p) {
		
//		System.out.println("Started adding pixel");
		
		// Reset any pixel list we've made before
		if (pixels != null) pixels = null;
		
		// Check if we're in the same file, if not then we're
		// not going to merge
		if (p.file() != file) {
			return false;
		}
		
		// If this is an RGB file check if the hue matches
		int hueDiff = 0;
		if (file.isRGB()) {
			hueDiff = p.h()-getAverageHue();
			while (hueDiff < -180) hueDiff += 360;
			while (hueDiff > 180) hueDiff -= 360;
		
			if (Math.abs(hueDiff) > collection.getPrefs().hueFlexibility()) {
				// Hue is too far off
//				System.out.println("Hue "+p.h()+" is too far from "+startingHue);
				return false;
			}
		}
		
		// Now we need to see if we're close enough to an existing pixel
		
		// We can be 1 slice out and up to 2 pixels away to be included.
		// Since everything gets added in slice order we never need to
		// look after the current slice
		//
		// Slices are numbered from 1, but the array is 0 indexed hence the
		// -2 and -1.
		
		for (int s=p.slice()-2;s<=p.slice()-1;s++) {
//			System.out.println("Slice is "+s);
			if (s<0) continue;
			if (s>=outerPixels.length) return false;
			
			// We can be up to 2 pixels away so we need to check from -2 to +2
			// of the current pixels position.
			
			for (int x=p.x()-2;x<=p.x()+2;x++) {
//				System.out.println("X is "+x);
				if (x<0) continue;
				if (x>=outerPixels[s].length) continue;
				
				// Now check if we're inside the range.
				
				// If there's nothing in this line then there's no point
				// checking
				if (outerPixels[s][x][0] == null) {
//					System.out.println("No existing data in slice "+s+" x:"+x);
					continue;
				}
				
				// If something's here already we need to be between the
				// existing limits (+/- 2).
				if (p.y() >= (outerPixels[s][x][0].y()-2) && p.y() <= (outerPixels[s][x][1].y()+2)) {
//					System.out.println("We're in");
					// We're in!

					++pixelCount;
					totalHueDeviation += hueDiff;
					totalRed += p.r() & 0xff;
					totalGreen += p.g() & 0xff;
					totalBlue += p.b() & 0xff;
					if (p.l() > brightestPixel.l()) brightestPixel = p;
					
					// Add ourselves
					if (outerPixels[p.slice()-1][p.x()][0] == null) {
						// We're the first ones in this slice 
						outerPixels[p.slice()-1][p.x()][0] = p;
						outerPixels[p.slice()-1][p.x()][1] = p;
					}
					else {
						if (p.y() < outerPixels[p.slice()-1][p.x()][0].y()){	
//							System.out.println("We're a new min pixel");
							outerPixels[p.slice()-1][p.x()][0] = p;
						}
						if (p.y() > outerPixels[p.slice()-1][p.x()][1].y()) {
//							System.out.println("We're a new max pixel");
							outerPixels[p.slice()-1][p.x()][1] = p;
						}
					}
				
					return true;
				}
				else {
//					System.out.println("Position "+p.y()+" not within "+(outerPixels[s][x][0].y()-2)+"-"+(outerPixels[s][x][1].y()+2));
				}
			}
		}
		
//		System.out.println("Rejecting pixel");
		
		return false;
	}
	
	public boolean mergeWith (SpatialGroup s) {
		// This checks whether we can merge ourselves with another spatial group
		
		// For a merge to happen we need to have a similar colour to the other
		// group and have a spatial overlap between our top and the other bottom
		// or our bottom and the other top.
		
		// First check the files match
		if (s.file != file) {
			return false;
		}
		
		// If this is an RGB file check the hue
		if (file.isRGB()) {
			int hueDiff = getAverageHue()-s.getAverageHue();
			while(hueDiff<-180) hueDiff +=360;
			while(hueDiff>180)hueDiff -=360;
			if (Math.abs(hueDiff)>collection.getPrefs().hueFlexibility()) {
//				System.out.println("Hue diff "+hueDiff+" is too big, not merging");
				// No need to go any further
				return false;
			}
		}
		
		// Now check for a spatial overlap
		int ourFirst = firstSlice();
		int ourLast = lastSlice();
		int theirFirst = s.firstSlice();
		int theirLast = s.lastSlice();

		int ourComparedSlice = 0;
		int theirComparedSlice = 0;

		
		// If one of the last slices falls between first and last then we're OK
		if (ourLast < theirLast && ourLast >=theirFirst) {
			ourComparedSlice = ourLast-1;
			theirComparedSlice = ourLast;
		}
		else if (theirLast < ourLast && theirLast >=ourFirst) {
			theirComparedSlice = theirLast-1;
			ourComparedSlice = theirLast;			
		}
		else {
//			System.out.println("No overlap between "+ourFirst+"-"+ourLast+" and "+theirFirst+"-"+theirLast);
			return false;
		}
				
		// Now we need to step through the pixels in these slices seeing
		// if we could fit any of them into the other group.
		
		for (int x=0;x<outerPixels[0].length;x++) {
			
			// Now check if we're inside the range.
			
			// If there's nothing in this line then there's no point
			// checking
			if (outerPixels[ourComparedSlice][x][0] == null || s.outerPixels[theirComparedSlice][x][0] == null) {
				continue;
			}

			// Since they both have some data let's look for an overlap
			if (outerPixels[ourComparedSlice][x][0].y()<=s.outerPixels[theirComparedSlice][x][1].y() && outerPixels[ourComparedSlice][x][1].y()>=s.outerPixels[theirComparedSlice][x][0].y()){
				// We can merge
				performSpatialGroupMerge(s);
				return true;
			}
		}			
		return false;
	}
	
	private void performSpatialGroupMerge (SpatialGroup sg) {
		// This method is not to be called lightly!  It will merge the contents
		// of s into this group.  It does no checking as to whether this merge is
		// valid, so be sure it is before you call this.
		
		pixels = null;
		pixelCount += sg.pixelCount;
		totalRed += sg.totalRed;
		totalGreen += sg.totalGreen;
		totalBlue += sg.totalBlue;
		if (sg.brightestPixel.l() > brightestPixel.l()) brightestPixel = sg.brightestPixel;
		
		int hueDeviation = startingHue - sg.startingHue;
		while (hueDeviation < -180) hueDeviation += 360;
		while (hueDeviation > 180) hueDeviation -= 360;
		totalHueDeviation += (sg.totalHueDeviation+hueDeviation);
		
		// We now need to step through the outer pixels working out if
		// we can expand our range.
		for (int s=0;s<outerPixels.length;s++) {
			for (int x=0;x<outerPixels[s].length;x++) {
				if (outerPixels[s][x][0] == null) {
					if (sg.outerPixels[s][x][0] != null) {
						outerPixels[s][x][0] = sg.outerPixels[s][x][0];
						outerPixels[s][x][1] = sg.outerPixels[s][x][1];
					}
				}
				else {
					if (sg.outerPixels[s][x][0] == null) continue;
					if (sg.outerPixels[s][x][0].y()<outerPixels[s][x][0].y())
						outerPixels[s][x][0] = sg.outerPixels[s][x][0];
					if (sg.outerPixels[s][x][1].y()>outerPixels[s][x][1].y())
						outerPixels[s][x][1] = sg.outerPixels[s][x][1];
				}
			}
		}
		
	}
		
	public StackMeasureFile getFile () {
		return file;
	}
	
	public int getPixelCount () {
		// This count is the full number of pixels accepted, not
		// just the outer pixel count.
		return pixelCount;
	}
	
	public Color getBrightestColor () {
		if (file.isRGB()) {
			return new Color(brightestPixel.r() & 0xff,brightestPixel.g() & 0xff,brightestPixel.b() & 0xff);
		}
		else {
			return file.getColour();
		}
		
	}
	
	public int [] getExtentForSlice (int slice) {
		int maxx = 0;
		int maxy = 0;
		int minx = 0;
		int miny = 0;
		boolean started = false;
		
		for (int x=0;x<outerPixels[slice-1].length;x++) {
			if (outerPixels[slice-1][x][0] == null) continue;
			if (started) {
				if (x > maxx) maxx = x;
				if (outerPixels[slice-1][x][1].y() > maxy) maxy = outerPixels[slice-1][x][1].y();
				if (x < minx) minx = x;
				if (outerPixels[slice-1][x][0].y() < miny) miny = outerPixels[slice-1][x][0].y();
			}
			else {
				maxx = x;
				maxy = outerPixels[slice-1][x][0].y();
				minx = x;
				miny = outerPixels[slice-1][x][0].y();
				started = true;
			}
		}
		
		if (started) {
			return new int [] {minx,miny,maxx,maxy};
		}
		else {
			return null; // No hits in this slice
		}
	}
	
	public Color getAverageColor () {
		if (file.isRGB()) {
			return new Color(totalRed/pixelCount,totalGreen/pixelCount,totalBlue/pixelCount);
		}
		else {
			return file.getColour();
		}
	}
	
	public int getAverageHue () {
		
		if (file.isRGB()) {
			int averageHue = startingHue+(totalHueDeviation/pixelCount);
			while (averageHue<0) averageHue+=360;
			while (averageHue>360) averageHue-=360;
			return averageHue;
		}
		else {
			return startingHue;
		}
	}
		
	public int firstSlice () {
		
		for (int s=0;s<outerPixels.length;s++) {
			for (int x=0;x<outerPixels[s].length;x++) {
				if (outerPixels[s][x][0]!=null) {
					return s+1;
				}
			}
		}
		
		return -1;
	}

	public int lastSlice () {
		for (int s=outerPixels.length-1;s>=0;s--) {
			for (int x=0;x<outerPixels[s].length;x++) {
				if (outerPixels[s][x][0]!=null) {
					return s+1;
				}
			}
		}
		
		return -1;
	}
	
	public int middleSlice () {
		return firstSlice()+((lastSlice()-firstSlice())/2);
	}

	public DistanceMeasure getClosestDistance (SpatialGroup g2) {
				
		DistanceMeasure closest = null;
		
		Pixel [] pixels1 = getPixels();
		Pixel [] pixels2 = g2.getPixels();

		for (int p1=0;p1<pixels1.length;p1++) {
			for (int p2=0;p2<pixels2.length;p2++) {
				// Take a shortcut if we're being compared to ourselves.
				if (pixels1[p1] == pixels2[p2]) {
					return new DistanceMeasure(pixels1[p1],pixels2[p2],0);
				}
				
				double dist = Math.sqrt(Math.pow((pixels1[p1].x()-pixels2[p2].x())*collection.getPrefs().pixelDistance(),2) + Math.pow((pixels1[p1].y()-pixels2[p2].y())*collection.getPrefs().pixelDistance(),2) + Math.pow((pixels1[p1].slice()-pixels2[p2].slice())*collection.getPrefs().stackDistance(),2));
				if (closest ==null || dist < closest.distance()) {
					closest = new DistanceMeasure(pixels1[p1],pixels2[p2],dist);
				}
			}
		}
		
		return closest;
	}
	
	public DistanceMeasure getCentralDistance (SpatialGroup g2) {
		Pixel p1 = getCentralPixel();
		Pixel p2 = g2.getCentralPixel();
		
		double dist = Math.sqrt(Math.pow((p1.x()-p2.x())*collection.getPrefs().pixelDistance(),2) + Math.pow((p1.y()-p2.y())*collection.getPrefs().pixelDistance(),2) + Math.pow((p1.slice()-p2.slice())*collection.getPrefs().stackDistance(),2));
		
		return new DistanceMeasure(p1,p2,dist);
	}
	
	protected Pixel getCentralPixel () {
		int x=0;
		int y=0;
		int slice = 0;
		
		Pixel [] p = getPixels();
		for (int i=0;i<p.length;i++) {
			x+=p[i].x();
			y+=p[i].y();
			slice+=p[i].slice();
		}
		x/=p.length;
		y/=p.length;
		slice/=p.length;
		
		// Since this is a virtual pixel we're only worried about position, not colour.
		return new Pixel(null,slice,x,y);
	}
	
	
	public Pixel [] getPixels () {

		// See if we've got a set ready to go
		if (pixels != null) return pixels;
		
		Vector v = new Vector();
		for (int s=0;s<outerPixels.length;s++) {
			for (int x=0;x<outerPixels[s].length;x++) {
				if (outerPixels[s][x][0] == null) continue;
				for (int y=0;y<outerPixels[s][x].length;y++) {
					v.add(outerPixels[s][x][y]);
				}
			}
		}

		pixels = (Pixel [])v.toArray(new Pixel [0]);
		
		return pixels;
	}
	
	public double getVolume () {
		return Math.pow(collection.getPrefs().pixelDistance(),2)*collection.getPrefs().stackDistance()*getPixelCount();
	}
	
	public int compareTo(Object g2) {
		// We make the comparison on the basis of the hue.
		
		int hDiff = getAverageHue() - ((SpatialGroup)g2).getAverageHue();
				
		while (hDiff > 180) hDiff -=360;
		while (hDiff < -180) hDiff +=360;
		return hDiff;
		
	}

}