/**
 * Copyright Copyright 2007-8 Simon Andrews
 *
 *    This file is part of StackMeasure.
 *
 *    StackMeasure is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    StackMeasure is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with StackMeasure; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
import ij.IJ;
import ij.ImagePlus;

import java.awt.Color;
import java.io.File;


public class StackMeasureFile {

	
	private static final Color [] defaultColours = new Color [] {Color.RED,Color.GREEN,Color.BLUE,Color.ORANGE,Color.MAGENTA,Color.CYAN,Color.PINK};
	private static int nextColourIndex = 0;
	
	private boolean isRGB;
	private File file;
	private Color colour;
	private ImagePlus ip;
	
	public StackMeasureFile (File file) {
		if (file == null) {
			throw new IllegalArgumentException("File can't be null");
		}
		if ((!file.exists()) ||(!file.canRead()) ) {
			throw new IllegalArgumentException("File doesn't exist, or couldn't be read");
		}
		this.file = file;
		
		ip = IJ.openImage(file.getAbsolutePath());
		
		if (ip == null) {
			throw new IllegalArgumentException("File was not a valid image file");
		}
		
		// We need to check if this is an RGB or black and white image.
		// RGB images get processed through the full pipeline to separate
		// the colours.  Black and white images get assigned a specific
		// colour which will never merge with anything else.

		if (ip.getBitDepth() == 24) {
			isRGB = true;
		}
		else {
			isRGB = false;
			colour = StackMeasureFile.defaultColours[StackMeasureFile.nextColourIndex];
			StackMeasureFile.nextColourIndex++;
			if (StackMeasureFile.nextColourIndex >= StackMeasureFile.defaultColours.length) {
				StackMeasureFile.nextColourIndex = 0;
			}
		}
	}
	
	public StackMeasureFile (File file, ImagePlus ip) {
		// In some cases it's useful to do the extraction of the imageplus object
		// outside this class.   This is particularly for file formats which allow
		// multiple stacks in one file (lsm for example).
		this.file = file;
		this.ip = ip;
		
		if (ip.getBitDepth() == 24) {
			isRGB = true;
		}
		else {
			isRGB = false;
			colour = StackMeasureFile.defaultColours[StackMeasureFile.nextColourIndex];
			StackMeasureFile.nextColourIndex++;
			if (StackMeasureFile.nextColourIndex >= StackMeasureFile.defaultColours.length) {
				StackMeasureFile.nextColourIndex = 0;
			}
		}

	}
	
	public static void resetColourIndexes() {
		nextColourIndex = 0;
	}
	
	public String name () {
		return (file.getName());
	}
	
	public void setColour (Color c) {
		if (isRGB) {
			throw new IllegalArgumentException("Can't set a fixed colour for an RGB image");
		}
		if (c == null) {
			throw new IllegalArgumentException("Colour can't be null");
		}
		colour = c;
	}
	
	public String toString() {
		return name();
	}
	
	public Color getColour () {
		if (isRGB) {
			return null;
		}
		else {
			return colour;
		}
	}
	
	public boolean isRGB () {
		return isRGB;
	}
	
	public ImagePlus getImagePlus () {
		return ip;
	}
}
